use serenity::{
  builder::CreateApplicationCommand,
  client::Context,
  model::{
    application::interaction::application_command::ApplicationCommandInteraction,
    prelude::{command::CommandOptionType, interaction::application_command::CommandDataOptionValue},
  },
};

use skynet_discord_bot::{get_server_config, DataBase, Servers};
use sqlx::{Pool, Sqlite};

pub(crate) mod user {}

pub(crate) mod server {
  use serde::{Deserialize, Serialize};
  use serenity::model::id::GuildId;
  use sqlx::Error;
  // this is to managfe the server side of commands related to minecraft
  use super::*;
  use skynet_discord_bot::set_roles::get_server_member_bulk;
  use skynet_discord_bot::{is_admin, Config};

  pub fn register(command: &mut CreateApplicationCommand) -> &mut CreateApplicationCommand {
    command
      .name("add_minecraft")
      .description("Enable the bot for this discord")
      .create_option(|option| {
        option
          .name("server_id")
          .description("ID of the Minecraft server hosted by the Computer Society")
          .kind(CommandOptionType::String)
          .required(true)
      })
  }

  pub async fn run(command: &ApplicationCommandInteraction, ctx: &Context) -> String {
    // check if user has high enough permisssions
    if let Some(msg) = is_admin(command, ctx).await {
      return msg;
    }
    let g_id = match command.guild_id {
      None => return "Not in a server".to_string(),
      Some(x) => x,
    };

    let server_minecraft = if let CommandDataOptionValue::String(id) = command
      .data
      .options
      .get(0)
      .expect("Expected server_id option")
      .resolved
      .as_ref()
      .expect("Expected server_id object")
    {
      Some(id.to_owned())
    } else {
      None
    };

    let db_lock = {
      let data_read = ctx.data.read().await;
      data_read.get::<DataBase>().expect("Expected Databse in TypeMap.").clone()
    };
    let db = db_lock.read().await;

    let server_data = match get_server_config(&db, &g_id).await {
      None => {
        return "No existing server config, have you used ``/add``?".to_string();
      }
      Some(mut x) => {
        x.server_minecraft = server_minecraft.clone();
        x
      }
    };

    match add_server(&db, &server_data).await {
      Ok(_) => {}
      Err(e) => {
        println!("{:?}", e);
        return format!("Failure to insert into Servers {:?}", server_data);
      }
    }

    let config_lock = {
      let data_read = ctx.data.read().await;
      data_read.get::<Config>().expect("Expected Config in TypeMap.").clone()
    };
    let config = config_lock.read().await;

    update_server(server_minecraft, &db, &g_id, &config).await;

    "Added/Updated minecraft_server info".to_string()
  }

  async fn add_server(db: &Pool<Sqlite>, server: &Servers) -> Result<Option<Servers>, Error> {
    sqlx::query_as::<_, Servers>(
      "
      UPDATE servers
      SET server_minecraft = ?2
      WHERE server = ?1; 
    ",
    )
    .bind(*server.server.as_u64() as i64)
    .bind(&server.server_minecraft)
    .fetch_optional(db)
    .await
  }

  /**
  loop through all members of server
  get a list of folks with mc accounts that are members
  and a list that arent members
  */
  async fn update_server(server_minecraft: Option<String>, db: &Pool<Sqlite>, g_id: &GuildId, config: &Config) {
    if let Some(server_id) = server_minecraft {
      let mut usernames = vec![];
      for member in get_server_member_bulk(db, g_id).await {
        if let Some(x) = member.minecraft {
          usernames.push(x);
        }
      }
      if !usernames.is_empty() {
        update_whitelist(&usernames, &server_id, &config.discord_minecraft).await;
      }
    }
  }

  pub async fn update_whitelist(add: &Vec<String>, server: &str, token: &str) {
    let url_base = format!("http://panel.games.skynet.ie/api/client/servers/{server}");
    let bearer = format!("Bearer {token}");

    async fn post<T: Serialize>(url: &str, bearer: &str, data: &T) {
      match surf::post(url)
        .header("Authorization", bearer)
        .header("Content-Type", "application/json")
        .header("Accept", "Application/vnd.pterodactyl.v1+json")
        .body_json(&data)
      {
        Ok(req) => {
          req.await.ok();
        }
        Err(e) => {
          dbg!(e);
        }
      }
    }

    #[derive(Deserialize, Serialize, Debug)]
    struct BodyCommand {
      command: String,
    }

    #[derive(Deserialize, Serialize, Debug)]
    struct BodyDelete {
      root: String,
      files: Vec<String>,
    }

    // delete whitelist
    let deletion = BodyDelete {
      root: "/".to_string(),
      files: vec!["whitelist.json".to_string()],
    };
    post(&format!("{url_base}/files/delete"), &bearer, &deletion).await;

    // reload the whitelist
    let data = BodyCommand {
      command: "whitelist reload".to_string(),
    };
    post(&format!("{url_base}/command"), &bearer, &data).await;

    for name in add {
      let data = BodyCommand {
        command: format!("whitelist add {name}"),
      };
      post(&format!("{url_base}/command"), &bearer, &data).await;
    }
  }
}
